import { EOL } from 'os';
import { basename } from 'path';
import camelcase from 'camelcase';
import { SourceNode } from '../library/source-map/index.js';
import { getRelativePath } from './index.js';
/**
 * Get .d.ts file path.
 * @param filePath The path to the source file (i.e. `/dir/foo.css`). It is absolute.
 * @returns The path to the .d.ts file. It is absolute.
 */
export function getDtsFilePath(filePath) {
    return filePath + '.d.ts';
}
function dashesCamelCase(str) {
    return str.replace(/-+(\w)/g, function (match, firstLetter) {
        return firstLetter.toUpperCase();
    });
}
function formatTokens(tokens, localsConvention) {
    const result = [];
    for (const token of tokens) {
        if (localsConvention === 'camelCaseOnly') {
            result.push({ ...token, name: camelcase(token.name) });
        }
        else if (localsConvention === 'camelCase') {
            result.push(token);
            result.push({ ...token, name: camelcase(token.name) });
        }
        else if (localsConvention === 'dashesOnly') {
            result.push({ ...token, name: dashesCamelCase(token.name) });
        }
        else if (localsConvention === 'dashes') {
            result.push(token);
            result.push({ ...token, name: dashesCamelCase(token.name) });
        }
        else {
            result.push(token); // asIs
        }
    }
    return result;
}
function generateTokenDeclarations(filePath, sourceMapFilePath, tokens, dtsFormatOptions, isExternalFile) {
    const formattedTokens = formatTokens(tokens, dtsFormatOptions?.localsConvention);
    const result = [];
    for (const token of formattedTokens) {
        // Only one original position can be associated with one generated position.
        // This is due to the sourcemap specification. Therefore, we output multiple type definitions
        // with the same name and assign a separate original position to each.
        for (let originalLocation of token.originalLocations) {
            if (originalLocation.filePath === undefined) {
                // If the original location is not specified, fallback to the source file.
                originalLocation = {
                    filePath: filePath,
                    start: { line: 1, column: 1 },
                    end: { line: 1, column: 1 },
                };
            }
            result.push(originalLocation.filePath === filePath || isExternalFile(originalLocation.filePath)
                ? new SourceNode(null, null, null, [
                    '& Readonly<{ ',
                    new SourceNode(originalLocation.start.line ?? null, 
                    // The SourceNode's column is 0-based, but the originalLocation's column is 1-based.
                    originalLocation.start.column - 1 ?? null, getRelativePath(sourceMapFilePath, originalLocation.filePath), `"${token.name}"`, token.name),
                    ': string }>',
                ])
                : // Imported tokens in non-external files are typed by dynamic import.
                    // See https://github.com/mizdra/happy-css-modules/issues/106.
                    new SourceNode(null, null, null, [
                        '& Readonly<Pick<(typeof import(',
                        `"${getRelativePath(filePath, originalLocation.filePath)}"`,
                        '))["default"], ',
                        `"${token.name}"`,
                        '>>',
                    ]));
        }
    }
    return result;
}
export function generateDtsContentWithSourceMap(filePath, dtsFilePath, sourceMapFilePath, tokens, dtsFormatOptions, isExternalFile) {
    const tokenDeclarations = generateTokenDeclarations(filePath, sourceMapFilePath, tokens, dtsFormatOptions, isExternalFile);
    let sourceNode;
    if (!tokenDeclarations || !tokenDeclarations.length) {
        sourceNode = new SourceNode(null, null, null, '');
    }
    else {
        sourceNode = new SourceNode(1, 0, getRelativePath(sourceMapFilePath, filePath), [
            'declare const styles:' + EOL,
            ...tokenDeclarations.map((tokenDeclaration) => ['  ', tokenDeclaration, EOL]),
            ';' + EOL,
            'export default styles;' + EOL,
        ]);
    }
    const codeWithSourceMap = sourceNode.toStringWithSourceMap({
        // Since sourcemap and type definitions are in the same directory, they can be referenced by relative paths.
        file: basename(dtsFilePath),
        sourceRoot: '',
    });
    return {
        dtsContent: codeWithSourceMap.code,
        sourceMap: codeWithSourceMap.map,
    };
}
//# sourceMappingURL=dts.js.map