import { resolve, relative } from 'path';
import * as process from 'process';
import * as util from 'util';
import { createCache } from '@file-cache/core';
import { createNpmPackageKey } from '@file-cache/npm';
import chalk from 'chalk';
import * as chokidar from 'chokidar';
import _glob from 'glob';
import { isGeneratedFilesExist, emitGeneratedFiles } from './emitter/index.js';
import { Locator } from './locator/index.js';
import { createDefaultResolver } from './resolver/index.js';
import { createDefaultTransformer } from './transformer/index.js';
import { getInstalledPeerDependencies, isMatchByGlob } from './util.js';
const glob = util.promisify(_glob);
export async function run(options) {
    const cwd = options.cwd ?? process.cwd();
    const silent = options.silent ?? false;
    const resolver = options.resolver ??
        createDefaultResolver({
            cwd,
            sassLoadPaths: options.sassLoadPaths,
            lessIncludePaths: options.lessIncludePaths,
            webpackResolveAlias: options.webpackResolveAlias,
        });
    const transformer = options.transformer ?? createDefaultTransformer({ cwd, postcssConfig: options.postcssConfig });
    const installedPeerDependencies = await getInstalledPeerDependencies();
    const cache = await createCache({
        mode: options.cacheStrategy ?? 'content',
        keys: [
            () => createNpmPackageKey(['happy-css-modules', ...installedPeerDependencies]),
            () => {
                return JSON.stringify(options);
            },
        ],
        noCache: !(options.cache ?? true),
    });
    const locator = new Locator({ transformer, resolver });
    const isExternalFile = (filePath) => {
        return !isMatchByGlob(filePath, options.pattern, { cwd });
    };
    async function processFile(filePath) {
        try {
            const result = await locator.load(filePath);
            await emitGeneratedFiles({
                filePath,
                tokens: result.tokens,
                emitDeclarationMap: options.declarationMap,
                dtsFormatOptions: {
                    localsConvention: options.localsConvention,
                },
                isExternalFile,
            });
            if (!silent)
                console.log(`${chalk.green(relative(cwd, filePath))} (generated)`);
        }
        catch (error) {
            if (error instanceof Error) {
                // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                console.error(chalk.red('[Error] ' + error.stack));
            }
            else {
                // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
                console.error(chalk.red('[Error] ' + error));
            }
            throw error;
        }
    }
    async function isChangedFile(filePath) {
        const result = await cache.getAndUpdateCache(filePath);
        if (result.error)
            throw result.error;
        return result.changed;
    }
    if (options.watch) {
        if (!silent)
            console.log('Watch ' + options.pattern + '...');
        const watcher = chokidar.watch([options.pattern.replace(/\\/g, '/')], { cwd });
        watcher.on('all', (eventName, filePath) => {
            if (eventName === 'add' || eventName === 'change') {
                processFile(resolve(cwd, filePath)).catch(() => {
                    // TODO: Emit a error by `Watcher#onerror`
                });
            }
        });
        return { close: async () => watcher.close() };
    }
    else {
        const filePaths = (await glob(options.pattern, { dot: true, cwd }))
            // convert relative path to absolute path
            .map((file) => resolve(cwd, file));
        const errors = [];
        for (const filePath of filePaths) {
            try {
                const _isGeneratedFilesExist = await isGeneratedFilesExist(filePath, options.declarationMap);
                const _isChangedFile = await isChangedFile(filePath);
                // Generate .d.ts and .d.ts.map only when the file has been updated.
                // However, if .d.ts or .d.ts.map has not yet been generated, always generate.
                if (!_isGeneratedFilesExist || _isChangedFile) {
                    await processFile(filePath);
                }
                else {
                    if (!silent)
                        console.log(chalk.gray(`${relative(cwd, filePath)} (skipped)`));
                }
            }
            catch (e) {
                errors.push(e);
            }
        }
        if (errors.length > 0)
            throw new AggregateError(errors, 'Failed to process files');
    }
    // Write cache state to file for persistence
    await cache.reconcile();
}
//# sourceMappingURL=runner.js.map