import { readFile, rm, writeFile } from 'fs/promises';
import { randomUUID } from 'node:crypto';
import { createRequire } from 'node:module';
import { dirname, join, resolve } from 'path';
import { fileURLToPath } from 'url';
import * as fileCacheNpm from '@file-cache/npm';
import { jest } from '@jest/globals';
import chalk from 'chalk';
import dedent from 'dedent';
import type { Watcher } from './runner.js';
import { createFixtures, exists, getFixturePath, waitForAsyncTask } from './test/util.js';

const require = createRequire(import.meta.url);

jest.unstable_mockModule('@file-cache/npm', () => ({
  ...fileCacheNpm, // Inherit native functions
  createNpmPackageKey: () => 'mocked-key',
}));

const { run } = await import('./runner.js');

// eslint-disable-next-line @typescript-eslint/no-empty-function
const consoleLogSpy = jest.spyOn(console, 'log').mockImplementation(() => {});
// eslint-disable-next-line @typescript-eslint/no-empty-function
const consoleErrorSpy = jest.spyOn(console, 'error').mockImplementation(() => {});

const defaultOptions = {
  pattern: 'test/**/*.{css,scss}',
  declarationMap: true,
  silent: true,
  cwd: getFixturePath('/'),
  cache: false,
};

const dir = join(dirname(fileURLToPath(import.meta.url)));

beforeEach(async () => {
  consoleLogSpy.mockClear();
  consoleErrorSpy.mockClear();
  await rm(resolve(dir, '../node_modules/.cache/happy-css-modules'), { recursive: true, force: true }); // clear cache
});

// Exit the watcher even if the test fails
let watcher: Watcher | undefined;
afterEach(async () => {
  if (watcher) {
    await watcher.close();
  }
});

test('generates .d.ts and .d.ts.map', async () => {
  createFixtures({
    '/test/1.css': '.a {}',
    '/test/2.css': '.b {}',
  });
  await run({ ...defaultOptions });
  expect(await readFile(getFixturePath('/test/1.css.d.ts'), 'utf8')).toMatchSnapshot();
  expect(await readFile(getFixturePath('/test/1.css.d.ts.map'), 'utf8')).toMatchSnapshot();
  expect(await readFile(getFixturePath('/test/2.css.d.ts'), 'utf8')).toMatchSnapshot();
  expect(await readFile(getFixturePath('/test/2.css.d.ts.map'), 'utf8')).toMatchSnapshot();
});

test('uses cache', async () => {
  createFixtures({
    '/test/1.css': '.a {}',
  });
  await run({ ...defaultOptions, declarationMap: true, silent: false, cache: true });
  expect(consoleLogSpy).toBeCalledTimes(1);
  expect(consoleLogSpy).toHaveBeenNthCalledWith(1, expect.stringContaining('generated'));
  consoleLogSpy.mockClear();

  // Skip generation
  await run({ ...defaultOptions, declarationMap: true, silent: false, cache: true });
  expect(consoleLogSpy).toBeCalledTimes(1);
  expect(consoleLogSpy).toHaveBeenNthCalledWith(1, expect.stringContaining('skipped'));
  consoleLogSpy.mockClear();

  // Generates if generated files are missing
  await rm(getFixturePath('/test/1.css.d.ts'));
  await run({ ...defaultOptions, declarationMap: true, silent: false, cache: true });
  expect(consoleLogSpy).toBeCalledTimes(1);
  expect(consoleLogSpy).toHaveBeenNthCalledWith(1, expect.stringContaining('generated'));
  consoleLogSpy.mockClear();

  // Generates if options are changed
  await run({ ...defaultOptions, declarationMap: false, silent: false, cache: true });
  expect(consoleLogSpy).toBeCalledTimes(1);
  expect(consoleLogSpy).toHaveBeenNthCalledWith(1, expect.stringContaining('generated'));
  consoleLogSpy.mockClear();
});

test('outputs logs', async () => {
  createFixtures({
    '/test/1.css': '.a {}',
  });
  await run({ ...defaultOptions, silent: false, cache: true });
  expect(consoleLogSpy).toBeCalledTimes(1);
  expect(consoleLogSpy).toHaveBeenNthCalledWith(1, `${chalk.green('test/1.css')} (generated)`);
  consoleLogSpy.mockClear();

  await run({ ...defaultOptions, silent: false, cache: true });
  expect(consoleLogSpy).toBeCalledTimes(1);
  expect(consoleLogSpy).toHaveBeenNthCalledWith(1, `${chalk.gray('test/1.css (skipped)')}`);
});

test.todo('changes dts format with localsConvention options');
test('does not emit declaration map if declarationMap is false', async () => {
  createFixtures({
    '/test/1.css': '.a {}',
  });
  await run({ ...defaultOptions, declarationMap: false });
  await expect(readFile(getFixturePath('/test/1.css.d.ts'), 'utf8')).resolves.not.toThrow();
  await expect(readFile(getFixturePath('/test/1.css.d.ts.map'), 'utf8')).rejects.toThrow(/ENOENT/);
});
test('supports transformer', async () => {
  createFixtures({
    '/test/1.scss': `.a { dummy: ''; }`,
  });
  await run({ ...defaultOptions });
  expect(await readFile(getFixturePath('/test/1.scss.d.ts'), 'utf8')).toMatchSnapshot();
  expect(await readFile(getFixturePath('/test/1.scss.d.ts.map'), 'utf8')).toMatchSnapshot();
});
test('watches for changes in files', async () => {
  createFixtures({
    '/test': {}, // empty directory
  });
  watcher = await run({ ...defaultOptions, watch: true });

  await writeFile(getFixturePath('/test/1.css'), '.a-1 {}');
  await waitForAsyncTask(500); // Wait until the file is written
  expect(await readFile(getFixturePath('/test/1.css.d.ts'), 'utf8')).toMatch(/a-1/);

  await writeFile(getFixturePath('/test/1.css'), '.a-2 {}');
  await waitForAsyncTask(500); // Wait until the file is written
  expect(await readFile(getFixturePath('/test/1.css.d.ts'), 'utf8')).toMatch(/a-2/);
});
test('returns an error if the file fails to process in non-watch mode', async () => {
  createFixtures({
    '/test/1.css': '.a {}',
    '/test/2.css': 'INVALID SYNTAX',
    '/test/3.css': 'INVALID SYNTAX',
  });
  const maybeError = await run({ ...defaultOptions, watch: false }).catch((e) => e);

  // The errors are aggregated into AggregateError.
  expect(maybeError).toBeInstanceOf(AggregateError);
  const error = maybeError as AggregateError;
  expect(error.message).toMatchInlineSnapshot(`"Failed to process files"`);
  expect(error.errors).toHaveLength(2);
  expect(error.errors[0]).toMatchInlineSnapshot(`<fixtures>/test/2.css:1:1: Unknown word`);
  expect(error.errors[1]).toMatchInlineSnapshot(`<fixtures>/test/3.css:1:1: Unknown word`);

  // The error is logged to console.error.
  expect(consoleErrorSpy).toHaveBeenCalledTimes(2);
  // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
  expect(consoleErrorSpy).toHaveBeenNthCalledWith(1, chalk.red('[Error] ' + error.errors[0]!.stack));
  // eslint-disable-next-line @typescript-eslint/restrict-plus-operands
  expect(consoleErrorSpy).toHaveBeenNthCalledWith(2, chalk.red('[Error] ' + error.errors[1].stack));

  // The valid files are emitted.
  expect(await exists(getFixturePath('/test/1.css.d.ts'))).toBe(true);
  expect(await exists(getFixturePath('/test/1.css.d.ts.map'))).toBe(true);
});
describe('handles external files', () => {
  beforeEach(() => {
    createFixtures({
      '/test/1.css': dedent`
      @import './2.css';
      @import 'external-library';
      .a {}
      `,
      '/test/2.css': `.b {}`,
      '/node_modules/external-library/index.css': `.c {}`,
    });
  });
  test('do not emit .dts for external files', async () => {
    await run({ ...defaultOptions });
    expect(await exists(getFixturePath('/test/1.css.d.ts'))).toBe(true);
    expect(await exists(getFixturePath('/test/2.css.d.ts'))).toBe(true);
    expect(await exists(getFixturePath('/node_modules/external-library/index.css.d.ts'))).toBe(false);
  });
  test('treats imported tokens from external files the same as local tokens', async () => {
    await run({ ...defaultOptions });
    expect(await readFile(getFixturePath('/test/1.css.d.ts'), 'utf8')).toMatchInlineSnapshot(`
          "declare const styles:
            & Readonly<Pick<(typeof import("./2.css"))["default"], "b">>
            & Readonly<{ "c": string }>
            & Readonly<{ "a": string }>
          ;
          export default styles;
          //# sourceMappingURL=./1.css.d.ts.map
          "
      `);
  });
});

test('sassLoadPaths', async () => {
  const sassLoadPaths = ['test/relative'];
  createFixtures({
    '/test/1.scss': dedent`
    @import '2.scss';
    `,
    '/test/relative/2.scss': `.a { dummy: ''; }`,
  });
  await run({ ...defaultOptions, sassLoadPaths }); // not throw
});

test('lessIncludePaths', async () => {
  const lessIncludePaths = ['test/relative'];
  createFixtures({
    '/test/1.less': dedent`
    @import '2.less';
    `,
    '/test/relative/2.less': `.a { dummy: ''; }`,
  });
  await run({ ...defaultOptions, lessIncludePaths }); // not throw
});

test('webpackResolveAlias', async () => {
  const webpackResolveAlias = { '@relative': 'test/relative' };
  createFixtures({
    '/test/1.less': dedent`
    @import '@relative/2.less';
    `,
    '/test/relative/2.less': `.a { dummy: ''; }`,
  });
  await run({ ...defaultOptions, webpackResolveAlias }); // not throw
});

test('postcssConfig', async () => {
  const uuid = randomUUID();
  const postcssConfig = `${uuid}/postcss.config.js`;
  createFixtures({
    [`/${uuid}/postcss.config.js`]: dedent`
    module.exports = {
      plugins: [
        require('${require.resolve('postcss-simple-vars')}'),
      ],
    };
    `,
    '/test/1.css': dedent`
    $prefix: foo;
    .$(prefix)_bar {}
    `,
  });
  await run({ ...defaultOptions, postcssConfig }); // not throw
});
